import re
from typing import Callable, List, TYPE_CHECKING, Optional, cast

if TYPE_CHECKING:
    from botocraft.services import (
        Service,
        Cluster,
        ContainerInstance,
        Task,
        TaskManager,
        TaskDefinition
    )


# ---------
# Functions
# ---------

def extract_task_family_and_revision(task_definition_arn: str) -> str:
    """
    Extract the task family and revision from a task definition ARN.

    Args:
        task_definition_arn: The ARN of the task definition.

    Returns:
        The task family and revision in the format ``<family>:<revision>``.
    """
    task_definition_arn_re = r"arn:aws:ecs:[^:]+:[^:]+:task-definition/(?P<family>[^:]+):(?P<revision>[0-9]+)"
    match = re.match(task_definition_arn_re, task_definition_arn)
    assert match, f'Could not extract task family and revision from {task_definition_arn}'
    return f'{match.group("family")}:{match.group("revision")}'


# ----------
# Decorators
# ----------


# Service

def ecs_services_only(func: Callable[..., List[str]]) -> Callable[..., List["Service"]]:
    """
    This wraps :py:meth:`botocraft.services.ecs.ServiceManager.list` to return a list of
    :py:class:`botocraft.services.ecs.Service` objects instead of only a list of
    ARNs.
    """
    def wrapper(self, *args, **kwargs) -> List["Service"]:
        arns = func(self, *args, **kwargs)
        services = []
        # We have to do this in batches of 10 because the get_many method,
        # which uses the boto3 ``describe_services`` method, only accepts 10 ARNs
        # at a time.
        for i in range(0, len(arns), 10):
            services.extend(
                self.get_many(
                    arns[i:i + 10],
                    cluster=kwargs['cluster'],
                    include=['TAGS']
                )
            )
        return services
    return wrapper


# Cluster

def ecs_clusters_only(func: Callable[..., List[str]]) -> Callable[..., List["Cluster"]]:
    """
    This wraps :py:meth:`botocraft.services.ecs.ClusterManager.list` to return a list of
    :py:class:`botocraft.services.ecs.Cluster` objects instead of only a list of
    ARNs.
    """
    def wrapper(self, *args, **kwargs) -> List["Cluster"]:
        arns = func(self, *args, **kwargs)
        clusters = []
        # We have to do this in batches of 100 because the get_many method,
        # which uses the boto3 ``describe_clusters`` method, only accepts 100 ARNs
        # at a time.
        for i in range(0, len(arns), 100):
            clusters.extend(
                self.get_many(clusters=arns[i:i + 100])
            )
        return clusters
    return wrapper


# TaskDefinition

def ecs_task_definitions_only(
    func: Callable[..., List[str]]
) -> Callable[..., List["TaskDefinition"]]:
    """
    Decorator to convert a list of ECS task definition identifiers to a list of
    :py:class:`botocraft.services.ecs.TaskDefinition` objects.
    """
    def wrapper(self, *args, **kwargs) -> List["TaskDefinition"]:
        identifiers = func(self, *args, **kwargs)
        task_definitions = []
        for identifier in identifiers:
            task_definitions.append(
                self.get(identifier, include=['TAGS'])
            )
        return task_definitions
    return wrapper


# ContainerInstance

def ecs_container_instances_only(
    func: Callable[..., List[str]]
) -> Callable[..., List["ContainerInstance"]]:
    """
    Decorator to convert a list of ECS container instance arns to a list of
    :py:class:`botocraft.services.ecs.ContainerInstance` objects.
    """
    def wrapper(self, *args, **kwargs) -> List["ContainerInstance"]:
        arns = func(self, *args, **kwargs)
        container_instances = []
        for i in range(0, len(arns), 100):
            container_instances.extend(
                self.get_many(
                    cluster=kwargs['cluster'],
                    containerInstances=arns[i:i + 100]
                )
            )
        return container_instances
    return wrapper


def ecs_container_instances_tasks_only(
    func: Callable[..., List[str]]
) -> Callable[..., List["Task"]]:
    """
    Decorator to convert a list of ECS container instance arns to a list of
    :py:class:`botocraft.services.ecs.ContainerInstance` objects.
    """
    def wrapper(self, *args, **kwargs) -> List["Task"]:
        from botocraft.services.ecs import Task  # pylint: disable=import-outside-toplevel
        arns = func(self, *args, **kwargs)
        tasks = []
        for i in range(0, len(arns), 100):
            tasks.extend(cast("TaskManager", Task.objects).get_many(arns[i:i + 100]))
        return tasks
    return wrapper


# Task
def ecs_task_populate_taskDefinition(
    func: Callable[..., Optional["Task"]]
) -> Callable[..., Optional["Task"]]:
    """
    This wraps :py:meth:`botocraft.services.ecs.TaskManager.get` to populate the
    :py:attr:`botocraft.services.ecs.Task.taskDefinition` attribute.

    We set the ``taskDefinition`` attribute to the task family and revision in the
    format ``<family>:<revision>``.  ``taskDefinition`` is an extra field that we
    add to the :py:class:`botocraft.services.ecs.Task` object that is not in the
    original botocore shape, but is useful for our purposes.
    """
    def wrapper(self, *args, **kwargs) -> Optional["Task"]:
        task = func(self, *args, **kwargs)
        if task:
            task.taskDefinition = extract_task_family_and_revision(task.taskDefinitionArn)
        return task
    return wrapper


def ecs_task_populate_taskDefinitions(
    func: Callable[..., List["Task"]]
) -> Callable[..., List["Task"]]:
    """
    This wraps :py:meth:`botocraft.services.ecs.TaskManager.get_many` to
    populate the :py:attr:`botocraft.services.ecs.Task.taskDefinition` attribute
    on each task.

    We set the ``taskDefinition`` attribute to the task family and revision in the
    format ``<family>:<revision>``.  ``taskDefinition`` is an extra field that we
    add to the :py:class:`botocraft.services.ecs.Task` object that is not in the
    original botocore shape, but is useful for our purposes.
    """
    def wrapper(self, *args, **kwargs) -> List["Task"]:
        tasks = func(self, *args, **kwargs)
        for task in tasks:
            task.taskDefinition = extract_task_family_and_revision(task.taskDefinitionArn)
        return tasks
    return wrapper


def ecs_tasks_only(
    func: Callable[..., List[str]]
) -> Callable[..., List["Task"]]:
    """
    Wrap :py:meth:`botocraft.services.ecs.TaskManager.list` to return a list of
    :py:class:`botocraft.services.ecs.Task` objects instead of only a list of
    ARNs.
    """
    def wrapper(self, *args, **kwargs) -> List["Task"]:
        arns = func(self, *args, **kwargs)
        tasks = []
        # We have to do this in batches of 100 because the get_many method,
        # which uses the boto3 ``describe_tasks`` method, only accepts 100 ARNs
        # at a time.
        for i in range(0, len(arns), 100):
            tasks.extend(
                self.get_many(
                    cluster=kwargs['cluster'],
                    tasks=arns[i:i + 100]
                )
            )
        return tasks
    return wrapper


# Mixins

class ECSServiceModelMixin:
    """
    This is a mixin for :py:class:`botocraft.services.ecs.Service` that adds
    some additional methods that we can't auto generate.
    """

    @property
    def required_cpu(self) -> int:
        """
        The required CPU for the service in CPU shares.  One full CPU is
        equivalent to 1024 CPU shares.
        """
        cpu: int = 0
        td = self.task_definition
        if td.cpu:
            cpu = int(td.cpu)
        else:
            for container in td.containerDefinitions:
                if container.cpu:
                    cpu += container.cpu
        return cpu

    @property
    def required_memory(self) -> int:
        """
        Return the required memory for the service in MiB.
        """
        memory: int = 0
        td = self.task_definition
        if td.memory:
            memory = int(td.memory)
        else:
            for container in td.containerDefinitions:
                if container.memory:
                    memory += container.memory
        return int(memory)

    @property
    def container_instances(self) -> List["ContainerInstance"]:
        """
        Return the :py:class:`botocraft.services.ecs.ContainerInstance` objects which
        are running our tasks for the service.
        """
        return [task.container_instance for task in self.tasks]

    @property
    def is_stable(self) -> bool:
        """
        Return whether the service is stable or not.
        """
        # this is the same test that the `services_stable` waiter uses
        return len(self.deployments) == 1 and (self.runningCount == self.desiredCount)

    def wait_until_stable(
        self,
        max_attempts: int = 40,
        delay: int = 15
    ) -> None:
        """
        Wait until the service is stable.

        Raises:
            botocore.exceptions.WaiterError: if the service is not stable after
                ``max_attempts``, or some other error occurred.

        Keyword Args:
            max_attempts: The maximum number of attempts to make before giving
                up.
            delay: The number of seconds to wait between attempts.
        """
        waiter_config = {}
        if max_attempts:
            waiter_config['maxAttempts'] = max_attempts
        if delay:
            waiter_config['delay'] = delay
        if waiter_config:
            waiter_config['operation'] = 'DescribeServices'
        waiter = self.objects.get_waiter(
            'services_stable',
            WaiterConfig=waiter_config
        )
        waiter.wait(
            cluster=self.clusterArn,
            services=[self.serviceName]
        )

    def scale(
        self,
        desired_count: int,
        wait: bool = False,
    ) -> None:
        """
        Scale the service to the desired count.  If ``wait`` is True, this will
        wait for the service to reach the desired count using the ``services_stable``
        boto3 waiter.

        Args:
            desired_count: The number of tasks to run.

        Keyword Args:
            wait: If True, wait for the service to reach the desired count.
        """
        self.objects.partial_update(  # type: ignore[attr-defined]
            self.serviceName,  # type: ignore[attr-defined]
            cluster=self.clusterArn,  # type: ignore[attr-defined]
            desiredCount=desired_count
        )
        waiter = self.objects.get_waiter('services_stable')  # type: ignore[attr-defined]
        if wait:
            waiter.wait(
                cluster=self.clusterArn,  # type: ignore[attr-defined]
                services=[self.serviceName]  # type: ignore[attr-defined]
            )


class ECSContainerInstanceModelMixin:

    @property
    def free_cpu(self) -> int:
        """
        Return the free CPU shares on the container instance.  One full CPU is
        equivalent to 1024 CPU shares.
        """
        value: int = 0
        for resource in self.remainingResources:
            if resource.name == 'CPU':
                value = int(resource.integerValue)
        return value

    @property
    def free_ram(self) -> int:
        """
        Return the free RAM in MiB on the container instance.
        """
        value: int = 0
        for resource in self.remainingResources:
            if resource.name == 'MEMORY':
                value = int(resource.integerValue)
        return value



from collections import OrderedDict
from typing import cast, Literal

from .base import ManagerMethodGenerator


class GeneralMethodGenerator(ManagerMethodGenerator):

    method_name: str = 'general'

    def kwargs(self, location: Literal['method', 'operation'] = 'method') -> OrderedDict[str, str]:
        """
        Just in case this ends up being used for a method that can paginate,
        we'll exclude the pagination arguments.
        """
        args = super().kwargs(location=location)
        if self.client.can_paginate(self.boto3_name):
            _args: OrderedDict[str, str] = OrderedDict()
            for _arg, arg_type in super().kwargs(location=location).items():
                if _arg not in self.PAGINATOR_ARGS:
                    _args[_arg] = arg_type
            return _args
        return args

    @property
    def return_type(self) -> str:
        """
        For generic methods:

        * If :py:attr:`ManagerMethodDefinition.return_type` is set, use that.
        * If :py:attr:`ManagerMethodDefinition.response_attr` is set, infer the
          return type from the response attribute.
        * Otherwise, return the response class name

        Returns:
            The name of the return type class.
        """
        response_class_name = self.response_class
        if self.method_def.return_type:
            return self.method_def.return_type
        if self.response_attr is None:
            return f'"{response_class_name}"'
        if self.output_shape is not None:
            response_attr_shape = self.output_shape.members[cast(str, self.response_attr)]
        return_type = self.shape_converter.convert(response_attr_shape, quote=True)
        return return_type

    @property
    def body(self) -> str:
        """
        Generate the method body for a general method.

        Returns:
            The method body.
        """
        if self.client.can_paginate(self.boto3_name):
            code = f"""
        paginator = self.client.get_paginator('{self.boto3_name}')
        {self.operation_args}
        response_iterator = paginator.paginate(**{{k: v for k, v in args.items() if v is not None}})
        results: {self.return_type} = []
        for _response in response_iterator:
            response = {self.response_class}(**_response)
"""
            if self.response_attr is not None:
                code += f"""
            if response.{self.response_attr}:
                results.extend(response.{self.response_attr})
"""
            else:
                code += """
            if response is not None:
                try:
                    # Test whether the response is iterable
                    iter(response)
                except TypeError:
                    # If it not, append the response to the results list
                    results.append(response)  # type: ignore[arg-type]
                else:
                    # If it is, extend the results list with the response
                    results.extend(response)  # type: ignore[arg-type]
"""
            code += """
            else:
                break
        return results
"""
        else:
            code = f"""
        {self.operation_args}
        {self.operation_call}
"""
            if self.return_type == 'None' or self.return_type == '"None"':
                pass
            elif self.response_attr is None:
                code += """
        return response
"""
            else:
                code += f"""
        return response.{self.response_attr}
"""
        return code



from .create import CreateMethodGenerator

from ...models import MethodDocstringDefinition


class UpdateMethodGenerator(CreateMethodGenerator):

    method_name: str = 'update'

    @property
    def docstrings_def(self) -> MethodDocstringDefinition:
        """
        Return the docstring for the method.
        """
        docstrings_def = super().docstrings_def
        docstrings_def.args['model'] = f'The :py:class:`{self.model_name}` to update.'
        return docstrings_def



from typing import Optional, Any, ClassVar
import re
from pydantic import BaseModel, ConfigDict

import boto3


class TransformMixin:

    def transform(
        self,
        attribrute: str,
        transformer: Optional[str],
    ) -> Any:
        """
        Transform an attribute using a regular expression into something else
        before it is returned.

        .. important::
            This only makes sense for attributes that are strings.

        ``transformer`` is a regular expression that will be used to transform
        the value of the attribute.

        * If the attribute is ``None``, it will be returned verbatim.
        * If ``transformer`` is ``None``, the attribute will be returned verbatim.
        * If ``transformer`` has no named groups, the attribute will be replaced
          with the value of the first group.
        * If ``transformer`` has named groups, the attribute will be replaced
          with a dictionary of the named groups.

        Raises:
            ValueError: If the attribute does not exist on the model.
            RuntimeError: If the transformer fails to match the attribute value.

        Args:
            attribute: The attribute to transform.
            transformer: The regular expression to use to transform the attribute.

        Returns:
            The transformed attribute.
        """
        if not hasattr(self, attribrute):
            raise ValueError(f"Invalid attribute: {self.__class__.__name__}.{attribrute}")
        if transformer is None:
            return getattr(self, attribrute)
        value = getattr(self, attribrute)
        if value is None:
            return None
        if match := re.search(transformer, value):
            if match.groupdict():
                return match.groupdict()
            return match.group(1)
        else:
            raise RuntimeError(
                f"Transformer failed to match: transformer=r'{transformer}', value='{getattr(self, attribrute)}'"
            )


class Boto3Model(TransformMixin, BaseModel):
    """
    The base class for all boto3 models.
    """
    model_config = ConfigDict(validate_assignment=True)


class ReadonlyBoto3Model(Boto3Model):
    """
    The base class for all boto3 models that are readonly.
    """
    model_config = ConfigDict(
        frozen=True,
        validate_assignment=True
    )


class Boto3ModelManager(TransformMixin):

    #: The name of the boto3 service.  Example: ``ec2``, ``s3``, etc.
    service_name: str

    def __init__(self) -> None:
        #: The boto3 client for the AWS service
        self.client = boto3.client(self.service_name)  # type: ignore
        #: The boto3 session to use for this manager.
        self.session = boto3.session.Session()

    def using(self, session: boto3.session.Session) -> "Boto3ModelManager":
        """
        Use a different boto3 session for this manager.

        Args:
            session: The boto3 session to use.
        """
        self.session = session
        self.client = session.client(self.service_name)  # type: ignore
        return self

    def serialize(self, arg: Any) -> Any:
        """
        Some of our botocraft methods use :py:class:`Boto3Model` objects as
        arguments (e.g. ``create``, ``update``), but boto3 methods expect simple
        Python types.  This method will serialize the model into a set of types
        that boto3 will understand if it is a :py:class:`Boto3Model` object.

        While serializing, we always exclude ``None`` values, because boto3
        doesn't like them.

        If ``arg`` is not a :py:class:`Boto3Model` object, it will be returned
        verbatim.

        Args:
            arg: the botocraft method argument to serialize.

        Returns:
            A properly serialized argument.
        """
        if arg is None:
            return None
        if isinstance(arg, Boto3Model):
            return arg.model_dump(exclude_none=True)
        elif isinstance(arg, list):
            # Oop, this is a list.  We need to serialize each item in the list.
            return [self.serialize(a) for a in arg]
        return arg

    def get(self, *args, **kwargs):
        raise NotImplementedError

    def list(self, *args, **kwargs):
        raise NotImplementedError

    def create(self, model, **kwargs):
        raise NotImplementedError

    def update(self, model, **kwargs):
        # Some models cannot be updated, so instead of raising a
        # NotImplementedError, we raise a RuntimeError.
        raise RuntimeError("This model cannot be updated.")

    def delete(self, pk: str, **kwargs):
        raise NotImplementedError

    def get_waiter(self, name: str) -> Any:
        """
        Get a boto3 waiter object for this service.

        Args:
            name: The name of the waiter to get.

        Returns:
            The boto3 waiter object.
        """
        return self.client.get_waiter(name)


class ReadonlyBoto3ModelManager(Boto3ModelManager):   # pylint: disable=abstract-method

    def create(self, model, **kwargs):
        raise RuntimeError("This model cannot be created.")

    def update(self, model, **kwargs):
        raise RuntimeError("This model cannot be updated.")

    def delete(self, pk: str, **kwargs):
        raise RuntimeError("This model cannot be deleted.")


class ModelIdentityMixin:

    @property
    def pk(self) -> Optional[str]:
        """
        Get the primary key of the model instance.

        Returns:
            The primary key of the model instance.
        """
        raise NotImplementedError

    @property
    def arn(self) -> Optional[str]:
        """
        Get the ARN of the model instance.

        Returns:
            The ARN of the model instance.
        """
        raise ValueError("The model does not have an ARN.")

    @property
    def name(self) -> Optional[str]:
        """
        Get the name of the model instance.

        Returns:
            The name of the model instance.
        """
        raise ValueError("The model does not have a name.")


class ReadonlyPrimaryBoto3Model(  # pylint: disable=abstract-method
    ModelIdentityMixin,
    ReadonlyBoto3Model
):

    #: The manager for this model
    objects: ClassVar[Boto3ModelManager]

    def save(self, **kwargs):
        """
        Save the model.
        """
        raise RuntimeError("Cannot save a readonly model.")

    def delete(self):
        """
        Delete the model.
        """
        raise RuntimeError("Cannot delete a readonly model.")


class PrimaryBoto3Model(  # pylint: disable=abstract-method
    ModelIdentityMixin,
    Boto3Model
):
    """
    The base class for all boto3 models that get returned as the primary object
    from a boto3 operation.
    """

    #: The manager for this model
    manager: ClassVar[Boto3ModelManager]

    def save(self, **kwargs):
        """
        Save the model.
        """
        if self.pk:
            return self.objects.update(self, **kwargs)
        return self.objects.create(self, **kwargs)

    def delete(self):
        """
        Delete the model.
        """
        if not self.pk:
            raise ValueError("Cannot delete a model that has not been saved.")
        return self.manager.delete(self.pk)



from typing import Optional, Type, Dict, cast, TYPE_CHECKING

import botocore.model

if TYPE_CHECKING:
    from botocraft.sync.service import ServiceGenerator, ModelGenerator


class AbstractShapeConverter:
    """
    The base class for all shape converters.
    """

    def __init__(
        self,
        shape_converter: "PythonTypeShapeConverter"
    ) -> None:
        self.service_generator = shape_converter.service_generator
        self.model_generator = shape_converter.model_generator
        self.shape_converter = shape_converter

    def model_exists(self, model_name: str) -> bool:
        """
        If we've already generated this model, return ``True``,
        else return ``False``.

        Args:
            model_name: the name of the model to check
        """
        model_def = self.model_generator.get_model_def(model_name)
        if model_def.alternate_name:
            model_name = model_def.alternate_name
        if model_name in self.model_generator.classes:
            # We generated this model earlier in this phase of service
            # generation
            return True
        if model_name in self.service_generator.classes:
            # We generated generated this model in a previous phase of service
            # generation
            return True
        if model_name in self.service_generator.service_def.models:
            # This model is belongs to this service, but we haven't generated it
            # yet.  Do so now.
            return False
        if model_name in self.service_generator.interface.models:
            # models we've generated so far in other services
            return True
        # This is some dependent model for this service. We'll generate it now.
        return False

    def import_line(self, model_name: str) -> Optional[str]:
        """
        Given ``model_name``, determine whether we've already
        generated this model somewhere else in the service, and
        """
        if model_name in self.service_generator.service_def.models:
            # This model is belongs to this service, but we haven't generated it
            # yet.  We don't need an import line, because the model lives
            # in this service.
            return None
        if model_name in self.service_generator.interface.models:
            import_path = self.service_generator.interface.models[model_name]
            return f"from {import_path} import {model_name}"
        return None

    def to_python(
        self,
        shape: botocore.model.Shape,
        quote: bool = False,
        name_only: bool = False
    ) -> str:
        """
        Converts the given shape ``shape`` to a python type,
        possibly using the model generator to generate
        pydantic models for nested structures.

        Args:
            shape: the botocore shape to convert

        Keyword Args:
            quote: if ``True``, we will quote the string in the
                return value.  This is used when we're generating
                manager methods, because the manager classes appear
                in the file before the models, so we need to quote
                the model names to avoid mypy errors.
            name_only: if ``True``, we will return the name of the model
                and not create the model in the model generator, if this
                is a structure shape.

        Returns:
            the python type for the shape
        """
        raise NotImplementedError


class StringShapeConverter(AbstractShapeConverter):

    def to_python(
        self,
        shape: botocore.model.Shape,
        quote: bool = False,
        name_only: bool = False
    ) -> str:
        if shape.type_name == 'string' or shape.name == 'String':
            value = cast(botocore.model.StringShape, shape)
            if value.enum:
                contents = ', '.join([f"'{value}'" for value in value.enum])
                python_type = f'Literal[{contents}]'
            else:
                python_type = 'str'
            return python_type
        raise ValueError(f'Not string type: {shape.type_name}')


class BooleanShapeConverter(AbstractShapeConverter):

    def to_python(
        self,
        shape: botocore.model.Shape,
        quote: bool = False,
        name_only: bool = False
    ) -> str:
        if shape.type_name == 'boolean':
            return 'bool'
        raise ValueError(f'Not boolean type: {shape.type_name}')


class IntegerShapeConverter(AbstractShapeConverter):

    def to_python(
        self,
        shape: botocore.model.Shape,
        quote: bool = False,
        name_only: bool = False
    ) -> str:
        if shape.type_name in ['integer', 'long']:
            return 'int'
        raise ValueError(f'Not integer type: {shape.type_name}')


class DoubleShapeConverter(AbstractShapeConverter):

    def to_python(
        self,
        shape: botocore.model.Shape,
        quote: bool = False,
        name_only: bool = False
    ) -> str:
        if shape.type_name == 'double':
            return 'float'
        raise ValueError(f'Not double type: {shape.type_name}')


class ListShapeConverter(AbstractShapeConverter):

    def to_python(
        self,
        shape: botocore.model.Shape,
        quote: bool = False,
        name_only: bool = False
    ) -> str:
        """
        Convert a list shape to a python type.  If the list members are not
        strings, we will generate a pydantic model for the list members

        Args:
            shape: the botocore shape to convert

        Keyword Args:
            quote: if ``True``, we will quote the string in the
                return value.  This is used when we're generating
                manager methods, because the manager classes appear
                in the file before the models, so we need to quote
                the model names to avoid mypy errors.
            name_only: if ``True``, we will return the name of the model
                and not create the model in the model generator, if this
                is a structure shape.

        Raises:
            ValueError: this is not a list shape

        Returns:
            a python type for the list
        """
        if shape.type_name != 'list':
            raise ValueError(f'Not list type: {shape.type_name}')
        element_shape = cast(botocore.model.ListShape, shape).member
        inner_model_name = self.shape_converter.convert(
            element_shape,
            quote=True,
            name_only=name_only
        )
        return f'List[{inner_model_name}]'


class MapShapeConverter(AbstractShapeConverter):

    def to_python(
        self,
        shape: botocore.model.Shape,
        quote: bool = False,
        name_only: bool = False
    ) -> str:
        """
        Convert a map shape to a python type.  We're assuming that
        the map key is always a string, and the value is always a
        string -- I haven't seen any other different examples.

        Args:
            shape: the botocore shape to convert

        Keyword Args:
            quote: if ``True``, we will quote the string in the
                return value.  This is used when we're generating
                manager methods, because the manager classes appear
                in the file before the models, so we need to quote
                the model names to avoid mypy errors.
            name_only: if ``True``, we will return the name of the model
                and not create the model in the model generator, if this
                is a structure shape.

        Raises:
            ValueError: this is not a map shape

        Returns:
            the python type for the map
        """
        if shape.type_name != 'map':
            raise ValueError(f'Not map type: {shape.type_name}')
        shape = cast(botocore.model.MapShape, shape)
        value_type = self.shape_converter.convert(
            shape.value,
            quote=True,
            name_only=name_only
        )
        key_type = self.shape_converter.convert(
            shape.key,
            quote=True,
            name_only=name_only
        )
        return f'Dict[{key_type}, {value_type}]'


class StructureShapeConverter(AbstractShapeConverter):

    def to_python(
        self,
        shape: botocore.model.Shape,
        quote: bool = False,
        name_only: bool = False
    ) -> str:
        """
        Convert a structure shape to a python type.  This
        ends up always being a pydantic model.

        Side Effects:
            Might add a new pydantic class to the service python module.

        Args:
            shape: the botocore shape to convert

        Keyword Args:
            quote: if ``True``, we will quote the string in the
                return value.  This is used when we're generating
                manager methods, because the manager classes appear
                in the file before the models, so we need to quote
                the model names to avoid mypy errors.
            name_only: if ``True``, we will return the name of the model
                and not create the model in the model generator.

        Raises:
            ValueError: this is not a structure shape

        Returns:
            the python type for the structure
        """
        if shape.type_name != 'structure':
            raise ValueError(f'Not structure type: {shape.type_name}')
        import_path: Optional[str] = None
        shape = cast(botocore.model.StructureShape, shape)
        model_def = self.model_generator.get_model_def(shape.name)
        model_name = shape.name
        if model_def.alternate_name:
            model_name = model_def.alternate_name
        if not name_only:
            if not self.model_exists(model_name):
                # We have not yet generated this model as a pydantic class
                model_name = self.model_generator.generate_model(shape.name, model_shape=shape)
            else:
                # We've already generated the model, so we just need to import it
                # if it's in another service, otherwise we can just use it

                # FIXME: we have a problem here that if we want to import the model
                # from another service, but the model has an alternate name, we
                # won't find it this way because ``model_name`` here will just be
                # ``shape.name``
                if import_path := self.import_line(model_name):
                    self.model_generator.imports.add(import_path)
            if quote and not import_path:
                return f'"{model_name}"'
        else:
            if quote:
                return f'"{model_name}"'
        return model_name


class TimestampShapeConverter(AbstractShapeConverter):

    def to_python(
        self,
        shape: botocore.model.Shape,
        quote: bool = False,
        name_only: bool = False
    ) -> str:
        """
        Convert a timestamp shape to a python type.   This
        ends up always being a ``datetime.datetime`` object.

        Side Effects:
            This adds the ``datetime`` import to the model generator.

        Args:
            shape: the botocore shape to convert

        Keyword Args:
            quote: if ``True``, we will quote the string in the
                return value.  This is used when we're generating
                manager methods, because the manager classes appear
                in the file before the models, so we need to quote
                the model names to avoid mypy errors.

        Raises:
            ValueError: this is not a timestamp shape

        Returns:
            the python type for the timestamp
        """
        if shape.type_name == 'timestamp':
            return 'datetime'
        self.model_generator.imports.add('from datetime import datetime')
        raise ValueError(f'Not timestamp type: {shape.type_name}')


class PythonTypeShapeConverter:

    CONVERTERS: Dict[str, Type[AbstractShapeConverter]] = {
        'string': StringShapeConverter,
        'boolean': BooleanShapeConverter,
        'integer': IntegerShapeConverter,
        'structure': StructureShapeConverter,
        'timestamp': TimestampShapeConverter,
        'double': DoubleShapeConverter,
        'list': ListShapeConverter,
        'map': MapShapeConverter,
    }

    def __init__(
        self,
        service_generator: "ServiceGenerator",
        model_generator: "ModelGenerator"
    ) -> None:
        #: The service generator.  We use this for accessing the
        #: model generator in case we need to build pydantic
        #: models out of a structure shape
        self.service_generator = service_generator
        self.model_generator = model_generator
        self.converters: Dict[str, AbstractShapeConverter] = {
            key: converter_class(self)
            for key, converter_class in self.CONVERTERS.items()
        }

    def convert(
        self,
        shape: botocore.model.Shape,
        quote: bool = False,
        name_only: bool = False
    ) -> str:
        for converter in self.converters.values():
            try:
                return converter.to_python(shape, quote=quote, name_only=name_only)
            except ValueError:
                pass
        raise ValueError(f'No converter for {shape.type_name}')
